#include <Adafruit_NeoPixel.h>
#include <Math.h>
#include <SPI.h>
#include <ESP8266WiFi.h>
#include <WiFiClient.h>
#include <ESP8266WebServer.h>
#include <EEPROM.h>
extern "C" {
#include "user_interface.h"
}

#define DEBUG
#ifdef DEBUG
#define DEBUG_PRINT(...)    Serial.print(__VA_ARGS__)
#define DEBUG_PRINTLN(...)  Serial.println(__VA_ARGS__)
#else
#define DEBUG_PRINT(...)
#define DEBUG_PRINTLN(...)
#endif

#define LED_PIN 0
#define BUTTON_PIN 2
#define RAINRANDOM 20
#define SNOWRANDOM 10
#define WEATHERCHECKINTERVAL 300000

#define SUNREDCOEFF   1
#define SUNGREENCOEFF 0.75

#define CLOUDREDCOEFF   0.25
#define CLOUDGREENCOEFF 0.25
#define CLOUDBLUECOEFF  0.25

#define SNOWREDCOEFF   0.19
#define SNOWGREENCOEFF 0.25
#define SNOWBLUECOEFF  0.25

#define ANIMATIONSPEED 100

#define headerHtml "<!DOCTYPE HTML><html><meta name='viewport' content='width=device-width, initial-scale=1.0'><head><style>table, th, td {border: 3px outset #545454;border-collapse:collapse;padding:10px;text-align:center;}</style></head><body style='background-color:#121211;color:#9896B3;font-family:Verdana,Geneva,sans-serif;font-weight:100;'><center><div style='display:table;position:absolute;height:100%;width:100%;text-align:center;'><div style='display:table-cell;vertical-align:middle;'><div style='margin-left:auto;margin-right:auto;width:90%;max-width:550px;'>"
#define footerHtml "</div></div></div></body></html>\n"

IPAddress myIp(1, 1, 1, 1);
IPAddress myGateway(1, 1, 1, 2);
IPAddress mySubnet(255, 255, 255, 0);

Adafruit_NeoPixel panel = Adafruit_NeoPixel(64, LED_PIN, NEO_GRB + NEO_KHZ800);

String zipCode = "19040";
String appId = "e698b87b47a2d62b37d6cae0eb93ef3e";

char urlhost[] = "api.openweathermap.org";

ESP8266WebServer server(80);
WiFiClient client;

enum weatherStates {
  THUNDERSTORM, //2xx
  DRIZZLE,      //3xx
  RAIN,         //5xx
  SNOW,         //6xx
  ATMOSPHERE,   //7xx
  CLEAR,        //800
  CLOUD,        //80x (x = 1-4)
  EXTREME,      //90x
  OTHERS        //9xx
};

enum weatherStates weatherNow = CLEAR;
enum weatherStates weather3hr = CLEAR;
enum weatherStates weather6hr = CLEAR;

void rain() {
  static byte drops[] = {0, 0, 0, 0, 0, 0, 0, 0};
  clearPanel();
  for (int i = 0; i < 8; i++) {
    if (drops[i] == 0) {
      int c = random(RAINRANDOM);
      if (!c) drops[i]++;
    }
    else {
      switch (++drops[i]) {
        case 0:
          break;
        case 1:
          panel.setPixelColor(63 - i * 8, 0, 0, 127);
          break;
        case 2:
          panel.setPixelColor(63 - i * 8, 0, 0, 127);
          panel.setPixelColor(63 - i * 8 - 1, 0, 0, 127);
          break;
        case 3:
          panel.setPixelColor(63 - i * 8, 0, 0, 63);
          panel.setPixelColor(63 - i * 8 - 1, 0, 0, 127);
          panel.setPixelColor(63 - i * 8 - 2, 0, 0, 127);
          break;
        case 4:
          panel.setPixelColor(63 - i * 8, 0, 0, 31);
          panel.setPixelColor(63 - i * 8 - 1, 0, 0, 63);
          panel.setPixelColor(63 - i * 8 - 2, 0, 0, 127);
          panel.setPixelColor(63 - i * 8 - 3, 0, 0, 127);
          break;
        case 5:
          panel.setPixelColor(63 - i * 8, 0, 0, 15);
          panel.setPixelColor(63 - i * 8 - 1, 0, 0, 31);
          panel.setPixelColor(63 - i * 8 - 2, 0, 0, 63);
          panel.setPixelColor(63 - i * 8 - 3, 0, 0, 127);
          panel.setPixelColor(63 - i * 8 - 4, 0, 0, 127);
          break;
        case 6:
          panel.setPixelColor(63 - i * 8, 0, 0, 7);
          panel.setPixelColor(63 - i * 8 - 1, 0, 0, 15);
          panel.setPixelColor(63 - i * 8 - 2, 0, 0, 31);
          panel.setPixelColor(63 - i * 8 - 3, 0, 0, 63);
          panel.setPixelColor(63 - i * 8 - 4, 0, 0, 127);
          panel.setPixelColor(63 - i * 8 - 5, 0, 0, 127);
          break;
        case 7:
          panel.setPixelColor(63 - i * 8, 0, 0, 3);
          panel.setPixelColor(63 - i * 8 - 1, 0, 0, 7);
          panel.setPixelColor(63 - i * 8 - 2, 0, 0, 15);
          panel.setPixelColor(63 - i * 8 - 3, 0, 0, 31);
          panel.setPixelColor(63 - i * 8 - 4, 0, 0, 63);
          panel.setPixelColor(63 - i * 8 - 5, 0, 0, 127);
          panel.setPixelColor(63 - i * 8 - 6, 0, 0, 127);
          break;
        case 8:
          panel.setPixelColor(63 - i * 8, 0, 0, 1);
          panel.setPixelColor(63 - i * 8 - 1, 0, 0, 3);
          panel.setPixelColor(63 - i * 8 - 2, 0, 0, 7);
          panel.setPixelColor(63 - i * 8 - 3, 0, 0, 15);
          panel.setPixelColor(63 - i * 8 - 4, 0, 0, 31);
          panel.setPixelColor(63 - i * 8 - 5, 0, 0, 63);
          panel.setPixelColor(63 - i * 8 - 6, 0, 0, 127);
          panel.setPixelColor(63 - i * 8 - 7, 0, 0, 127);
          break;
        case 9:
          panel.setPixelColor(63 - i * 8 - 1, 0, 0, 1);
          panel.setPixelColor(63 - i * 8 - 2, 0, 0, 3);
          panel.setPixelColor(63 - i * 8 - 3, 0, 0, 7);
          panel.setPixelColor(63 - i * 8 - 4, 0, 0, 15);
          panel.setPixelColor(63 - i * 8 - 5, 0, 0, 31);
          panel.setPixelColor(63 - i * 8 - 6, 0, 0, 63);
          panel.setPixelColor(63 - i * 8 - 7, 0, 0, 127);
          panel.setPixelColor(constrain(63 - i * 8 - 14, 0, 62), 0, 0, 32);
          panel.setPixelColor(constrain(63 - i * 8 + 2, 6, 63), 0, 0, 32);
          break;
        case 10:
          panel.setPixelColor(63 - i * 8 - 2, 0, 0, 1);
          panel.setPixelColor(63 - i * 8 - 3, 0, 0, 3);
          panel.setPixelColor(63 - i * 8 - 4, 0, 0, 7);
          panel.setPixelColor(63 - i * 8 - 5, 0, 0, 15);
          panel.setPixelColor(63 - i * 8 - 6, 0, 0, 31);
          panel.setPixelColor(63 - i * 8 - 7, 0, 0, 63);
          break;
        case 11:
          panel.setPixelColor(63 - i * 8 - 3, 0, 0, 1);
          panel.setPixelColor(63 - i * 8 - 4, 0, 0, 3);
          panel.setPixelColor(63 - i * 8 - 5, 0, 0, 7);
          panel.setPixelColor(63 - i * 8 - 6, 0, 0, 15);
          panel.setPixelColor(63 - i * 8 - 7, 0, 0, 31);
          break;
        case 12:
          panel.setPixelColor(63 - i * 8 - 4, 0, 0, 1);
          panel.setPixelColor(63 - i * 8 - 5, 0, 0, 3);
          panel.setPixelColor(63 - i * 8 - 6, 0, 0, 7);
          panel.setPixelColor(63 - i * 8 - 7, 0, 0, 15);
          break;
        case 13:
          panel.setPixelColor(63 - i * 8 - 5, 0, 0, 1);
          panel.setPixelColor(63 - i * 8 - 6, 0, 0, 3);
          panel.setPixelColor(63 - i * 8 - 7, 0, 0, 7);
          break;
        case 14:
          panel.setPixelColor(63 - i * 8 - 6, 0, 0, 1);
          panel.setPixelColor(63 - i * 8 - 7, 0, 0, 3);
          break;
        case 15:
          panel.setPixelColor(63 - i * 8 - 7, 0, 0, 1);
          drops[i] = 0;
          break;
      }
    }
  }
}
void sun() {
  static int frame = 0;
  clearPanel();
  float offset = (1 + sin((frame / 127.00) * 6.28)) / 3;
  float brightnessCenterRed =   (127 * offset + 127) * SUNREDCOEFF;
  float brightnessCenterGreen = (127 * offset + 127) * SUNGREENCOEFF;
  float brightness1layerRed =   (63 * offset + 63) * SUNREDCOEFF;
  float brightness1layerGreen = (63 * offset + 63) * SUNGREENCOEFF;
  float brightness2layerRed =   (31 * offset + 31) * SUNREDCOEFF;
  float brightness2layerGreen = (31 * offset + 31) * SUNGREENCOEFF;
  float brightness3layerRed =   (15 * offset + 15) * SUNREDCOEFF;
  float brightness3layerGreen = (15 * offset + 15) * SUNGREENCOEFF;
  panel.setPixelColor(0, brightness3layerRed, brightness3layerGreen, 0);
  panel.setPixelColor(1, brightness3layerRed, brightness3layerGreen, 0);
  panel.setPixelColor(2, brightness3layerRed, brightness3layerGreen, 0);
  panel.setPixelColor(3, brightness3layerRed, brightness3layerGreen, 0);
  panel.setPixelColor(4, brightness3layerRed, brightness3layerGreen, 0);
  panel.setPixelColor(5, brightness3layerRed, brightness3layerGreen, 0);
  panel.setPixelColor(6, brightness3layerRed, brightness3layerGreen, 0);
  panel.setPixelColor(7, brightness3layerRed, brightness3layerGreen, 0);
  panel.setPixelColor(8, brightness3layerRed, brightness3layerGreen, 0);

  panel.setPixelColor(9, brightness2layerRed, brightness2layerGreen, 0);
  panel.setPixelColor(10, brightness2layerRed, brightness2layerGreen, 0);
  panel.setPixelColor(11, brightness2layerRed, brightness2layerGreen, 0);
  panel.setPixelColor(12, brightness2layerRed, brightness2layerGreen, 0);
  panel.setPixelColor(13, brightness2layerRed, brightness2layerGreen, 0);
  panel.setPixelColor(14, brightness2layerRed, brightness2layerGreen, 0);

  panel.setPixelColor(15, brightness3layerRed, brightness3layerGreen, 0);
  panel.setPixelColor(16, brightness3layerRed, brightness3layerGreen, 0);

  panel.setPixelColor(17, brightness2layerRed, brightness2layerGreen, 0);
  panel.setPixelColor(18, brightness1layerRed, brightness1layerGreen, 0);
  panel.setPixelColor(19, brightness1layerRed, brightness1layerGreen, 0);
  panel.setPixelColor(20, brightness1layerRed, brightness1layerGreen, 0);
  panel.setPixelColor(21, brightness1layerRed, brightness1layerGreen, 0);
  panel.setPixelColor(22, brightness2layerRed, brightness2layerGreen, 0);

  panel.setPixelColor(23, brightness3layerRed, brightness3layerGreen, 0);
  panel.setPixelColor(24, brightness3layerRed, brightness3layerGreen, 0);

  panel.setPixelColor(25, brightness2layerRed, brightness2layerGreen, 0);
  panel.setPixelColor(26, brightness1layerRed, brightness1layerGreen, 0);
  panel.setPixelColor(27, brightnessCenterRed, brightnessCenterGreen, 0);
  panel.setPixelColor(28, brightnessCenterRed, brightnessCenterGreen, 0);
  panel.setPixelColor(29, brightness1layerRed, brightness1layerGreen, 0);
  panel.setPixelColor(30, brightness2layerRed, brightness2layerGreen, 0);

  panel.setPixelColor(31, brightness3layerRed, brightness3layerGreen, 0);
  panel.setPixelColor(32, brightness3layerRed, brightness3layerGreen, 0);

  panel.setPixelColor(33, brightness2layerRed, brightness2layerGreen, 0);
  panel.setPixelColor(34, brightness1layerRed, brightness1layerGreen, 0);
  panel.setPixelColor(35, brightnessCenterRed, brightnessCenterGreen, 0);
  panel.setPixelColor(36, brightnessCenterRed, brightnessCenterGreen, 0);
  panel.setPixelColor(37, brightness1layerRed, brightness1layerGreen, 0);
  panel.setPixelColor(38, brightness2layerRed, brightness2layerGreen, 0);

  panel.setPixelColor(39, brightness3layerRed, brightness3layerGreen, 0);
  panel.setPixelColor(40, brightness3layerRed, brightness3layerGreen, 0);

  panel.setPixelColor(41, brightness2layerRed, brightness2layerGreen, 0);
  panel.setPixelColor(42, brightness1layerRed, brightness1layerGreen, 0);
  panel.setPixelColor(43, brightness1layerRed, brightness1layerGreen, 0);
  panel.setPixelColor(44, brightness1layerRed, brightness1layerGreen, 0);
  panel.setPixelColor(45, brightness1layerRed, brightness1layerGreen, 0);
  panel.setPixelColor(46, brightness2layerRed, brightness2layerGreen, 0);

  panel.setPixelColor(47, brightness3layerRed, brightness3layerGreen, 0);
  panel.setPixelColor(48, brightness3layerRed, brightness3layerGreen, 0);

  panel.setPixelColor(49, brightness2layerRed, brightness2layerGreen, 0);
  panel.setPixelColor(50, brightness2layerRed, brightness2layerGreen, 0);
  panel.setPixelColor(51, brightness2layerRed, brightness2layerGreen, 0);
  panel.setPixelColor(52, brightness2layerRed, brightness2layerGreen, 0);
  panel.setPixelColor(53, brightness2layerRed, brightness2layerGreen, 0);
  panel.setPixelColor(54, brightness2layerRed, brightness2layerGreen, 0);

  panel.setPixelColor(55, brightness3layerRed, brightness3layerGreen, 0);
  panel.setPixelColor(56, brightness3layerRed, brightness3layerGreen, 0);
  panel.setPixelColor(57, brightness3layerRed, brightness3layerGreen, 0);
  panel.setPixelColor(58, brightness3layerRed, brightness3layerGreen, 0);
  panel.setPixelColor(59, brightness3layerRed, brightness3layerGreen, 0);
  panel.setPixelColor(60, brightness3layerRed, brightness3layerGreen, 0);
  panel.setPixelColor(61, brightness3layerRed, brightness3layerGreen, 0);
  panel.setPixelColor(62, brightness3layerRed, brightness3layerGreen, 0);
  panel.setPixelColor(63, brightness3layerRed, brightness3layerGreen, 0);
  frame = (frame + 1) % 127;
}
void cloud() {
  static byte current[64];
  static byte desired[64];
  clearPanel();
  for (int i = 0; i < 64; i++) {
    if (current[i] > desired[i]) current[i]--;
    else if (current[i] < desired[i]) current[i]++;
    else if (current[i] == desired[i]) desired[i] = random(255);
    panel.setPixelColor(i, (float)current[i]*CLOUDREDCOEFF, (float)current[i]*CLOUDGREENCOEFF, (float)current[i]*CLOUDBLUECOEFF);
  }
}
void snow() {
  static byte snowflakes[8] = {1, 5, 14, 20, 28, 45, 54, 60};
  static int frameskip = 0;
  if (frameskip > 4){
    clearPanel();
    for (int i = 0; i < 8; i++) {
      int c = random(5);
      switch (c) {
        case 0:
          snowflakes[i] = (snowflakes[i] - 8) % 64;
          break;
        case 1:
          snowflakes[i] = (snowflakes[i] - 9) % 64;
          break;
        case 2:
          snowflakes[i] = (snowflakes[i] - 1) % 64;
          break;
        case 3:
          snowflakes[i] = (snowflakes[i] + 7) % 64;
          break;
        case 4:
          snowflakes[i] = (snowflakes[i] + 8) % 64;
          break;
      }
      panel.setPixelColor((snowflakes[i] - 7) % 64, 16 * SNOWREDCOEFF, 16 * SNOWGREENCOEFF, 16 * SNOWBLUECOEFF);
      panel.setPixelColor((snowflakes[i] - 8) % 64, 16 * SNOWREDCOEFF, 16 * SNOWGREENCOEFF, 16 * SNOWBLUECOEFF);
      panel.setPixelColor((snowflakes[i] - 9) % 64, 16 * SNOWREDCOEFF, 16 * SNOWGREENCOEFF, 16 * SNOWBLUECOEFF);
      panel.setPixelColor((snowflakes[i] + 1) % 64, 16 * SNOWREDCOEFF, 16 * SNOWGREENCOEFF, 16 * SNOWBLUECOEFF);
      panel.setPixelColor((snowflakes[i] - 1) % 64, 16 * SNOWREDCOEFF, 16 * SNOWGREENCOEFF, 16 * SNOWBLUECOEFF);
      panel.setPixelColor((snowflakes[i] + 7) % 64, 16 * SNOWREDCOEFF, 16 * SNOWGREENCOEFF, 16 * SNOWBLUECOEFF);
      panel.setPixelColor((snowflakes[i] + 8) % 64, 16 * SNOWREDCOEFF, 16 * SNOWGREENCOEFF, 16 * SNOWBLUECOEFF);
      panel.setPixelColor((snowflakes[i] + 9) % 64, 16 * SNOWREDCOEFF, 16 * SNOWGREENCOEFF, 16 * SNOWBLUECOEFF);
  
      panel.setPixelColor(snowflakes[i], 128 * SNOWREDCOEFF, 128 * SNOWGREENCOEFF, 128 * SNOWBLUECOEFF);
    }
    frameskip = 0;
  }
  frameskip++;
}

void safeRestart() {
  DEBUG_PRINTLN("Restarting...");
  clearPanel();
  panel.show();
  ESP.restart();
  DEBUG_PRINTLN("WHAT ARE YOU DOING HERE");
  while (1) yield();
  DEBUG_PRINTLN("YOU HAVE MET WITH A TERRIBLE FATE, HAVEN'T YOU");
}
void clearSettings() {
  DEBUG_PRINTLN("Clearing Settings");
  for (int i = 0; i < 256; i++) EEPROM.write(i, 0);
  EEPROM.commit();
  DEBUG_PRINTLN("EEPROM cleared");
  safeRestart();
}
void clearPanel() {
  for (int i = 0; i < panel.numPixels(); i++) {
    panel.setPixelColor(i, 0, 0, 0);
  }
}
void requestWeather() {
  if (!client.connect(urlhost, 80)) {
    DEBUG_PRINTLN("connection failed");
  }
  else {
    String url = "/data/2.5/forecast?zip=";
    url += zipCode;
    url += ",us&APPID=";
    url += appId;
    url += "&cnt=3";
    DEBUG_PRINT("Requesting URL: ");
    DEBUG_PRINTLN(url);
    client.print(String("GET ") + url + " HTTP/1.1\r\n" + "Host: " + urlhost + "\r\n" + "Connection: close\r\n\r\n");
  }
}
void errorPage(byte reasonCode) {
  String s;
  DEBUG_PRINTLN("Error Page");
  s += "<span style='font-weight:600;'>Sorry, but I ran into an error during setup!</span><br><br>";
  switch (reasonCode) {
    case 0: s += "<span style='font-weight:300;'>SSID Error: Please make sure the SSID exists, is less than 16 characters long, and is a valid name.</span><br><br>";
    case 1: s += "<span style='font-weight:300;'>Password Error: Please make sure the password is valid and is less than 16 characters long.</span><br><br>";
    case 2: s += "<span style='font-weight:300;'>Zip Code Error: Please enter a valid 5-digit Zip Code.</span><br><br>";
    default: s += "<span style='font-weight:300;'>Unknown Error: Please email me and tell me what happened right before this error popped up: jarek@soniktech.com </span><br><br>";
  }
  s += "<br><br><input type='button' onClick='history.go(-1)' value='Back to Setup'>";
  s += footerHtml;
  server.send(200, "text/html", s);
  DEBUG_PRINTLN("Client disconnected");
}
void confirmPage() {
  DEBUG_PRINTLN("Got Post");
  String passString;
  String ssidString;
  String zipCodeString;
  String apiKeyString;
  String eepromString;

  if (server.arg("ssid").length() > 32) {
    errorPage(0);
    return;
  }
  else ssidString = server.arg("ssid");

  if (server.arg("pass").length() > 32) {
    errorPage(1);
    return;
  }
  else passString = server.arg("pass");

  if (server.arg("zipCode").length() > 5) {
    errorPage(2);
    return;
  }
  else zipCodeString = server.arg("zipCode");

  if (server.arg("apiKey").length() > 32) {
    errorPage(3);
    return;
  }
  if (server.arg("apiKey").length() < 32) {
    apiKeyString = "e698b87b47a2d62b37d6cae0eb93ef3e";
  }
  else apiKeyString = server.arg("apiKey");

  String s = headerHtml;
  s += "Thanks! I'm now going to connect to ";
  s += ssidString;
  if (passString != "") s += " using the password you have given me";
  s += ".<br><br>I will monitor the weather for zip code ";
  s += zipCodeString;
  s += " and use the appId ";
  s += apiKeyString;
  s += " to grab weather data.<br><br>";
  s += "You can clear my settings and rerun this setup by pressing the Clear button. This button clears all saved settings and returns me to a factory default state.";
  s += footerHtml;

  eepromString = "<ssid>";
  eepromString += ssidString;
  eepromString += "<pass>";
  eepromString += passString;
  eepromString += "<zip>";
  eepromString += zipCodeString;
  eepromString += "<api>";
  eepromString += apiKeyString;
  eepromString += "<end>";

  DEBUG_PRINTLN("Writing EEPROM");
  EEPROM.write(0, 42);
  for (int i = 0; i < (eepromString.length()); i++) EEPROM.write(i + 1, eepromString.charAt(i));
  EEPROM.write(eepromString.length() + 2, 42);
  EEPROM.commit();

  DEBUG_PRINT("EEPROM Written: ");
  for (int i = 0; i < (eepromString.length() + 1); i++) DEBUG_PRINT(char(EEPROM.read(i)));
  DEBUG_PRINTLN("");

  server.send(200, "text/html", s); // Send the response to the client
  DEBUG_PRINTLN("Client disconnected");  // The client will actually be disconnected when the function returns and 'client' object is detroyed
  yield();
  WiFi.softAPdisconnect(true);
  yield();
  safeRestart();
}
void setupPage() {
  int accessPoints = WiFi.scanNetworks();
  String s = headerHtml;
  DEBUG_PRINTLN("Access Points Found:");
  s += "<span style='font-weight:600;'>Hello, I will forecast the weather for you. Let's quickly set me up!</span><br><br>";
  s += "<form action='' method='POST'><table style='width:100%;'><tr><td>Here are the networks I can see,<br>which one can I connect to?</td><td>";
  if (accessPoints == 0) {
    s += "Can't Find Any Open WiFi Hotspots!";
    DEBUG_PRINTLN("No Access Points Found!");
  }
  else {
    s += "<input type='radio' name='ssid' value='";
    s += String(WiFi.SSID(0));
    s += "' checked>";
    s += String(WiFi.SSID(0));
    s += "<br>";
    DEBUG_PRINT("- ");
    DEBUG_PRINTLN(WiFi.SSID(0));
    for (int i = 1; i < accessPoints; i++) {
      s += "<input type='radio' name='ssid' value='";
      s += String(WiFi.SSID(i));
      s += "'>";
      s += String(WiFi.SSID(i));
      s += "<br>";
      DEBUG_PRINT("- ");
      DEBUG_PRINTLN(WiFi.SSID(i));
    }
  }
  s += "<br><br><input type='button' onClick='history.go(0)' value='Rescan Networks'></td></tr>";
  s += "<tr><td>If your WiFi Hotspot has a password,<br>please type it in here:</td><td><input type='password' style='text-align:center;' name='pass' value=''></td></tr>";
  s += "<tr><td>What ZIP Code should I look<br>up weather for?:</td><td><input type='number' style='text-align:center;' name='zipCode' value=''></td></tr>";
  s += "<tr><td>If you have an OpenWeather API Key<br>please type it in here:</td><td><input type='text' style='text-align:center;' name='apiKey' value=''></td></tr>";
  s += "<tr><td>Click Next to save settings</td><td><input type='submit' value='Next'></td></tr>";
  s += "</table></form>";
  s += footerHtml;
  server.send(200, "text/html", s); // Send the response to the client
  DEBUG_PRINTLN("Client disconnected");  // The client will actually be disconnected when the function returns and 'client' object is detroyed
}
void checkWeatherResponse() {
  while (client.available()) {
    String storedRequestString = client.readStringUntil('\r');
    if (storedRequestString.charAt(1) == '{') {
      storedRequestString.remove(0, 1);
      storedRequestString.remove(storedRequestString.length() - 1);
      DEBUG_PRINT("String: ");
      DEBUG_PRINTLN(storedRequestString);
      int index1 = storedRequestString.indexOf("\"dt\":");
      int index2 = storedRequestString.indexOf("\"dt\":", index1 + 1);
      int index3 = storedRequestString.indexOf("\"dt\":", index2 + 1);
      int weatherIndex1 = storedRequestString.indexOf("\"weather\":[{\"id\":", index1);
      int weatherIndex2 = storedRequestString.indexOf("\"weather\":[{\"id\":", index2);
      int weatherIndex3 = storedRequestString.indexOf("\"weather\":[{\"id\":", index3);
      String forecastString1 = storedRequestString.substring(weatherIndex1 + 17, weatherIndex1 + 20);
      String forecastString2 = storedRequestString.substring(weatherIndex2 + 17, weatherIndex2 + 20);
      String forecastString3 = storedRequestString.substring(weatherIndex3 + 17, weatherIndex3 + 20);
      DEBUG_PRINTLN(forecastString1);
      DEBUG_PRINTLN(forecastString2);
      DEBUG_PRINTLN(forecastString3);

      char weather1 = forecastString1.charAt(0);
      char weather11 = forecastString1.charAt(2);
      char weather2 = forecastString2.charAt(0);
      char weather22 = forecastString2.charAt(2);
      char weather3 = forecastString3.charAt(0);
      char weather33 = forecastString2.charAt(2);
      switch (weather1) {
        case 2: weatherNow = THUNDERSTORM; break;
        case 3: weatherNow = DRIZZLE; break;
        case 5: weatherNow = RAIN; break;
        case 6: weatherNow = SNOW; break;
        case 7: weatherNow = ATMOSPHERE; break;
        case 8: if (weather11 == '0') weatherNow = CLEAR; else weatherNow = CLOUD; break;
        case 9: weatherNow = OTHERS; break;
      }
      switch (weather2) {
        case 2: weather3hr = THUNDERSTORM; break;
        case 3: weather3hr = DRIZZLE; break;
        case 5: weather3hr = RAIN; break;
        case 6: weather3hr = SNOW; break;
        case 7: weather3hr = ATMOSPHERE; break;
        case 8: if (weather22 == '0') weather3hr = CLEAR; else weather3hr = CLOUD; break;
        case 9: weather3hr = OTHERS; break;
      }
      switch (weather3) {
        case 2: weather6hr = THUNDERSTORM; break;
        case 3: weather6hr = DRIZZLE; break;
        case 5: weather6hr = RAIN; break;
        case 6: weather6hr = SNOW; break;
        case 7: weather6hr = ATMOSPHERE; break;
        case 8: if (weather33 == '0') weather6hr = CLEAR; else weather6hr = CLOUD; break;
        case 9: weather6hr = OTHERS; break;
      }
    }
  }
}
void drawWeather() {
  //weatherNow = SNOW; //use this to test animation patterns
  switch (weatherNow) {
    case CLOUD:
      cloud();
      break;
    case RAIN:
      rain();
      break;
    case CLEAR:
      sun();
      break;
    case SNOW:
      snow();
      break;
    default:
      for (int i = 0; i < 64; i++) {
        panel.setPixelColor(i, 16, 0, 0);
      }
      break;
  }
  panel.show();
}
void setup() {
  randomSeed(analogRead(A0));
  WiFi.persistent(false);
  WiFi.mode(WIFI_STA);

  Serial.begin(115200);
  DEBUG_PRINTLN("Serial Online");

  EEPROM.begin(256);
  DEBUG_PRINTLN("EEPROM Online");

  pinMode(BUTTON_PIN, INPUT_PULLUP);
  attachInterrupt(BUTTON_PIN, clearSettings, FALLING);
  DEBUG_PRINTLN("Button Online");

  pinMode(LED_PIN, OUTPUT);
  panel.begin();
  clearPanel();
  for (int i = 0; i < 64; i++) {
    panel.setPixelColor(i, 16, 0, 0);
  }
  panel.show();
  DEBUG_PRINTLN("Pixels Online");

  String ssidString = "";
  String passString = "";
  String storedRequestString = "";
  DEBUG_PRINTLN("Strings Initialized");

  DEBUG_PRINT("EEPROM(0) = ");
  DEBUG_PRINTLN(EEPROM.read(0));

  DEBUG_PRINT("Button Status: ");
  DEBUG_PRINTLN(digitalRead(BUTTON_PIN));

  if ((EEPROM.read(0) != 42) || (digitalRead(BUTTON_PIN) == LOW)) {
    clearPanel();
    for (int i = 0; i < 64; i++) {
      panel.setPixelColor(i, 16, 16, 0);
    }
    panel.show();
    uint8_t mac[WL_MAC_ADDR_LENGTH];
    WiFi.softAPmacAddress(mac);
    DEBUG_PRINT("WiFi MAC : ");
    DEBUG_PRINTLN(WiFi.softAPmacAddress());

    server.begin();
    DEBUG_PRINTLN("Server Started");

    server.on("/", HTTP_POST, confirmPage);
    server.on("/", setupPage);
    DEBUG_PRINTLN("Server Pages Set, Waiting for Clients");

    WiFi.mode(WIFI_AP);
    WiFi.softAPConfig(myIp, myGateway, mySubnet);
    DEBUG_PRINTLN("WiFi AP Online");

    EEPROM.write(0, 0);
    EEPROM.commit();
    DEBUG_PRINTLN("Low, Clearing EEPROM");
    DEBUG_PRINTLN("EEPROM(0) != 42, EEPROM invalid");

    for (int i = 0; i < 256; ++i) EEPROM.write(i, 0);
    DEBUG_PRINTLN("EEPROM cleared");

    String macId = String(mac[WL_MAC_ADDR_LENGTH - 2], HEX) + String(mac[WL_MAC_ADDR_LENGTH - 1], HEX);
    macId.toUpperCase();
    String apString = "Weathr Setup " + macId;
    WiFi.softAP(apString.c_str());
    DEBUG_PRINT("WiFi SSID Set : ");
    DEBUG_PRINTLN(apString);

    while (1) {
      server.handleClient();
      yield();
    }
  }
  clearPanel();
  for (int i = 0; i < 64; i++) {
    panel.setPixelColor(i, 0, 0, 16);
  }
  panel.show();

  DEBUG_PRINTLN("EEPROM(0) = 42, using saved settings");

  for (int i = 0; i < 256; i++) storedRequestString += char(EEPROM.read(i));
  DEBUG_PRINT("EEPROM: ");
  DEBUG_PRINTLN(storedRequestString);

  int passIndex = storedRequestString.indexOf("<pass>");
  int zipCodeIndex = storedRequestString.indexOf("<zip>");
  int appIdIndex = storedRequestString.indexOf("<api>");
  int endIndex = storedRequestString.indexOf("<end>");

  ssidString = storedRequestString.substring(7, passIndex);
  passString = storedRequestString.substring(passIndex + 6, zipCodeIndex);
  zipCode = storedRequestString.substring(zipCodeIndex + 5, appIdIndex);
  appId = storedRequestString.substring(appIdIndex + 5, endIndex);

  WiFi.begin(ssidString.c_str(), passString.c_str());
  DEBUG_PRINT("Connecting to ");
  DEBUG_PRINTLN(ssidString);
  DEBUG_PRINT("Password ");
  DEBUG_PRINT(passString);

  int timeout = 0;
  while (WiFi.status() != WL_CONNECTED) {
    delay(1000);
    DEBUG_PRINT(WiFi.status());
    DEBUG_PRINT(".");
    timeout++;
    if (timeout > 60) {
      DEBUG_PRINTLN("Timeout Reached");
      EEPROM.write(0, 0);
      EEPROM.commit();
      DEBUG_PRINTLN("EEPROM Reset");
      safeRestart();
    }
  }

  clearPanel();
  for (int i = 0; i < 64; i++) {
    panel.setPixelColor(i, 0, 64, 0);
  }
  panel.show();

  DEBUG_PRINTLN("WiFi connected");
  DEBUG_PRINTLN("IP address: ");
  DEBUG_PRINTLN(WiFi.localIP());

  requestWeather();
}
void loop() {
  static unsigned long weatherTimer = 0;
  static unsigned long frameTimer = 0;
  if ((weatherTimer + WEATHERCHECKINTERVAL) < millis()) {
    requestWeather();
    weatherTimer = millis();
  }
  checkWeatherResponse();
  if ((frameTimer + ANIMATIONSPEED) < millis()) {
    drawWeather();
    frameTimer = millis();
  }
  yield();
}

